#!/usr/bin/python
##################################################################
##                                                              ##
## Author:	Louis-Philippe Lemieux Perreault                    ##
## Date:	18-12-2008                                          ##
## Source:	correct_pedcheck_errors                             ##
##                                                              ##
## Function: Corrects the errors found by pedcheck (level 1     ##
##           and level 2 errors) in a given pedfile             ##
##                                                              ##
##################################################################


import os
import sys
import shutil

from optparse import OptionParser



# The 'OptionParser' object and usage
parser = OptionParser()
parser.set_usage("Usage: python " + sys.argv[0] + " [OPTIONS]")



##
## The main method
##
def main():
	
	# Parse the command line options
	pedfileName, errorFileName, outputFileName, modify_outputFileName = parseOptions()
	
	mendelian_errors, outfile_suffix, nbErrors = parse_error_file(open(errorFileName, 'rb'))
	
	# The name of the output file is not specified, so give a default name (with the correct level of error)
	if modify_outputFileName:
		outputFileName += outfile_suffix
	
	print "These are the files that will be used:"
	print "\tpedfile:   ", pedfileName
	print "\terror file:", errorFileName
	print "\touput file:", outputFileName
	
	print "\nThere", ['is','are'][nbErrors > 1], nbErrors, "Mendelian error" + ['', 's'][nbErrors > 1]
	
	if nbErrors == 0:
		# There is no Mendelian error, so we only copy the file
		shutil.copyfile(pedfileName, outputFileName)
	else:
		# There are Mendelian errors, so we need to correct the file
		correct_pedfile(open(pedfileName, 'rb'), open(outputFileName, 'wb'), mendelian_errors)



##
## Corrects the pedfile by setting to 'undefined' (value 0) the
## nuclear family (father, mother and childs)
##
def correct_pedfile(inFile, outFile, errors):

	to_print = []

	# Reads each line of inFile (the pedfile)
	for line in inFile:
		data = line.strip().split(' ')
		
		# The first 2 columns are identifiers for our data structure (PED, ID)
		pedID, ID = data[:2]
		
		# If the key (pedID, ID) is in 'error', we need to correct all of the 'locus'
		if (pedID, ID) in errors:
			for locus in errors[(pedID, ID)]:
				# The 'locus' is the position in data ((2 * locus) + 3) and ((2 * locus) + 4)
				data[(2 * locus) + 3] = '0'
				data[(2 * locus) + 4] = '0'
		
		to_print.append(data)
	
	# Prints the new pedfile
	for line in to_print:
		print >>outFile, ' '.join(line)
						
	inFile.close()
	outFile.close()



##
## Parse the pedcheck error files in order to
## find the level 1 and level 2 errors
##
def parse_error_file(inFile):

	nbErrors = 0
	
	error_of_ID = {}
	outFileSuffix = ""

	line = inFile.readline()
	
	while line != "":
		if 'LEVEL 0 ERRORS' in line:
			# We have found the 'LEVEL 0 ERRORS' boundary
			inFile.readline()
			line = inFile.readline()
			
			if 'No Level 0 errors' not in line:
				# There is at least one Level 0 error
				# The user must correct it and re-run pedcheck
				print >>sys.stderr, "There is at least one 'Level 0 error' in the pedfile."
				print >>sys.stderr, "Manually correct the pedfile and re-run pedcheck."
				inFile.close()
				sys.exit(1)
		
		if 'LEVEL 1 ERRORS' in line:
			# We have found the 'LEVEL 1 ERRORS' boundary
			outFileSuffix = "_level_1"
			
			while line != "":
				line = inFile.readline()
				
				if 'GENOTYPE ERROR' in line:
					# We have found a level 1 error description
					nbErrors += 1
					data = line.strip().split(' ')
					
					while '' in data:
						data.remove('')
					error_locus = int(data[data.index('Locus') + 1])
					pedigreeID = data[data.index('Pedigree') + 1]

					# We need to skip till we have 'ORIGINAL SCORING' in the line
					# Then one more
					while 'ORIGINAL SCORING' not in line:
						line = inFile.readline()
					line = inFile.readline()
					
					data = line.strip().split(' ')
					
					while '' in data:
						data.remove('')

					IDs_to_correct = []
					IDs_to_correct.append(data[data.index('Father') + 1][:-1])		# The father
					IDs_to_correct.append(data[data.index('Mother') + 1][:-1])		# The mother
					
					line = inFile.readline()
					while 'Child' not in line:
						line = inFile.readline()
						
					while 'Child' in line:
						# We have a child
						data = line.strip().split(' ')
						
						while '' in data:
							data.remove('')
						
						IDs_to_correct.append(data[data.index('Child') + 1][:-1])	# All the childs
						
						line = inFile.readline()
					
					# Now, we need to populate the error map 'error_of_ID'
					for ID in IDs_to_correct:
						if (pedigreeID, ID) not in error_of_ID:
							error_of_ID[(pedigreeID, ID)] = [error_locus]
						else:
							error_of_ID[(pedigreeID, ID)].append(error_locus)
							
							
		elif 'LEVEL 2 ERRORS' in line:
			# We have found the 'LEVEL 1 ERRORS' boundary
			outFileSuffix = "_level_2"
			
			while line != "":
				line = inFile.readline()
				
				if 'GENOTYPE ERROR' in line:
					all_untyped = True
					
					# We have found a level 2 error description
					nbErrors += 1
					data = line.strip().split(' ')
					
					# Lets remove empty cells
					while '' in data:
						data.remove('')
						
					error_locus = int(data[data.index('Locus') + 1])
					pedigreeID = data[data.index('Pedigree') + 1]

					# We need to skip till we have 'Father' in the line
					while 'Father' not in line:
						line = inFile.readline()
					
					# Processing the 'father' line
					data = line.strip().split(' ')
					
					# Lets remove empty cells
					while '' in data:
						data.remove('')
					
					# Check if father is untyped
					if data[0] == '(T)':
						all_untyped = False

					IDs_to_correct = []
					IDs_to_correct.append(data[data.index('Father') + 1][:-1])		# The father
					
					# Processing the 'mother' line
					line = inFile.readline()
					
					# Find the 'mother' line
					while 'Mother' not in line:
						line = inFile.readline()
					
					data = line.strip().split(' ')
					
					while '' in data:
						data.remove('')
					
					# Check if mother is untyped
					if data[0] == '(T)':
						all_untyped = False
						
					IDs_to_correct.append(data[data.index('Mother') + 1][:-1])		# The mother
					
					# Lets find the 'Child' line
					line = inFile.readline()
					while 'Child' not in line:
						line = inFile.readline()
					
					while 'Child' in line:
						# While we have a child
						data = line.strip().split(' ')
						
						# Removing blank cells
						while '' in data:
							data.remove('')
						
						# Check if the child is untyped
						if data[0] == '(T)':
							all_untyped = False
						
						IDs_to_correct.append(data[data.index('Child') + 1][:-1])	# All the childs
						
						line = inFile.readline()
					
					if all_untyped:
						# All the individuals of this nuclear family are untyped...
						print >>sys.stderr, "********************** ERROR LEVEL 2 *******************************"
						print >>sys.stderr, "The pedcheck level 2 error file includes the subset of untyped"
						print >>sys.stderr, "individuals. Pedigree number is", pedigreeID
						print >>sys.stderr, "Please correct haplotypes of the individuals in the subset manually."
						print >>sys.stderr, "********************************************************************"
             			
						sys.exit(1)

					# Now, we need to populate the error map 'error_of_ID'
					for ID in IDs_to_correct:
						if (pedigreeID, ID) not in error_of_ID:
							error_of_ID[(pedigreeID, ID)] = [error_locus]
						else:
							error_of_ID[(pedigreeID, ID)].append(error_locus)
									
		line = inFile.readline()

	inFile.close()

	return error_of_ID, outFileSuffix, nbErrors



##
## Parse the options and arguments
##
def parseOptions():

	# Required arguments
	# The pedfile name
	parser.add_option("-p", "--pedfile", dest = "pedfileName",
						help = "The pedfile FILE name",
						metavar = "FILE")
	
	# The pedcheck error file name (default is the default name of pedcheck algorithm)
	parser.add_option("-e", "--errorFile", dest = "errorFileName", default="pedcheck.err",
						help = "The error FILE from pedcheck (DEFAULT is 'pedcheck.err')",
						metavar = "FILE")
						
	parser.add_option("-o", "--output", dest = "outputFileName",
						help = "The output FILE (DEFAULT is pedfile name with .clean_level_X)",
						metavar = "FILE")
	
	# Parse the option
	options, args = parser.parse_args()
	
	pedfileName = options.pedfileName
	errorFileName = options.errorFileName
	outputFileName = options.outputFileName
	
	# If no pedfile is specified or pedfile doesn't exists
	if pedfileName == None:
		usage("You must specify a pedfile")
	if not os.path.exists(pedfileName):
		usage("File '" + pedfileName + "' doesn't exists")
	
	# If pedcheck error file doesn't exists
	if not os.path.exists(errorFileName):
		usage("File '" + errorFileName + "' doesn't exists")
	
	# If the output file name is not specified, we give the pedfile name + prefix '.clean'.
	# In order to give a _level_X prefix, we need to parse the error file...
	if outputFileName == None:
		outputFileName = pedfileName + '.clean'
		modify_outputFileName = True
	else:
		modify_outputFileName = False
	
	return pedfileName, errorFileName, outputFileName, modify_outputFileName
	


##
## Prints an ERROR message and exit with code 1
##		
def usage(message):

	print >>sys.stderr, "\nERROR: " + message + '\n'
	parser.print_help()
	sys.exit(1)



if __name__ == '__main__':
	try:
		main()
		
	except KeyboardInterrupt:
		print >>sys.stderr, "\nExiting on user cancel"
